import os


class APIRepo:
    def __init__(self, url=None):
        default_url = "https://api.eotdl.com/"
        # default_url = "http://localhost:8010/"
        self.url = url if url else os.getenv("EOTDL_API_URL", default_url)

    def format_response(self, response):
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]



"""
Curation module
"""

from .stac.dataframe import STACDataFrame  # , read_stac
from .stac.stac import STACGenerator
from .stac.parsers import STACIdParser, StructuredParser, UnestructuredParser
from .stac.dataframe_labeling import UnlabeledStrategy, LabeledStrategy



"""
Sentinel-Hub data access module.
"""

from .client import SHClient
from .parameters import SHParameters, SH_PARAMETERS_DICT
from .evalscripts import EvalScripts
from .utils import evaluate_sentinel_parameters, imagery_from_tmp_to_dir



from pathlib import Path
import yaml

from ..auth import with_auth
from .metadata import Metadata
from ..repos import ModelsAPIRepo
from ..shared import calculate_checksum
from ..files import ingest_files


def ingest_model(path, verbose=False, logger=print):
    path = Path(path)
    if not path.is_dir():
        raise Exception("Path must be a folder")
    # if "catalog.json" in [f.name for f in path.iterdir()]:
    #     return ingest_stac(path / "catalog.json", logger)
    return ingest_folder(path, verbose, logger)


def retrieve_model(metadata, user):
    repo = ModelsAPIRepo()
    data, error = repo.retrieve_model(metadata.name)
    # print(data, error)
    if data and data["uid"] != user["sub"]:
        raise Exception("Model already exists.")
    if error and error == "Model doesn't exist":
        # create dataset
        data, error = repo.create_model(metadata.dict(), user["id_token"])
        # print(data, error)
        if error:
            raise Exception(error)
        data["id"] = data["model_id"]
    return data["id"]


@with_auth
def ingest_folder(folder, verbose=False, logger=print, user=None):
    repo = ModelsAPIRepo()
    # load metadata
    metadata = yaml.safe_load(open(folder.joinpath("metadata.yml"), "r").read()) or {}
    metadata = Metadata(**metadata)
    # retrieve model (create if doesn't exist)
    model_id = retrieve_model(metadata, user)
    # ingest files
    return ingest_files(
        repo, model_id, folder, verbose, logger, user, endpoint="models"
    )



"""
Module for STAC Asset Generators
"""

from os import remove, listdir
from os.path import dirname, join, basename, abspath

import pandas as pd
import rasterio
import pystac

from ...tools.metadata import remove_raster_metadata


MEDIA_TYPES_DICT = {
    "tif": pystac.MediaType.GEOTIFF,
    "tiff": pystac.MediaType.GEOTIFF,
    "png": pystac.MediaType.PNG,
    "jpg": pystac.MediaType.JPEG,
    "jpeg": pystac.MediaType.JPEG,
}


class STACAssetGenerator:
    """
    Standard STAC Asset Generator
    """

    type = "None"

    def __init__(self):
        pass

    @classmethod
    def extract_assets(cls, obj_info: pd.DataFrame):
        """
        Generate a single asset from the raster file

        :param raster_path: path to the raster file
        """
        # If there is no bands, create a single band asset from the file, assuming thats a singleband raster
        raster_path = obj_info["image"].values[0]
        title = basename(raster_path).split(".")[0]
        # Get the file extension
        raster_format = raster_path.split(".")[-1]
        asset = pystac.Asset(
            href=abspath(raster_path),
            title=title,
            media_type=MEDIA_TYPES_DICT[raster_format],
            roles=["data"],
        )

        return [asset]


class BandsAssetGenerator(STACAssetGenerator):
    """
    Bands STAC Asset Generator
    """

    type = "Bands"

    def __init__(self) -> None:
        super().__init__()

    def extract_assets(self, obj_info: pd.DataFrame):
        """
        Extract the assets from the raster file from the bands column

        :param raster_path: path to the raster file
        """
        asset_list = []
        # File path
        raster_path = obj_info["image"].values[0]
        # Bands
        bands = obj_info["bands"].values
        bands = bands[0] if bands else None

        if bands:
            with rasterio.open(raster_path, "r") as raster:
                if isinstance(bands, str):
                    bands = [bands]
                for band in bands:
                    i = bands.index(band)
                    raster_format = raster_path.split(".")[
                        -1
                    ]  # Will be used later to save the bands files
                    try:
                        single_band = raster.read(i + 1)
                    except IndexError:
                        single_band = raster.read(1)
                    band_name = f"{band}.{raster_format}"
                    output_band = join(dirname(raster_path), band_name)
                    # Copy the metadata
                    metadata = raster.meta.copy()
                    metadata.update({"count": 1})
                    # Write the band to the output folder
                    with rasterio.open(output_band, "w", **metadata) as dest:
                        dest.write(single_band, 1)
                    # Instantiate pystac asset and append it to the list
                    asset_list.append(
                        pystac.Asset(
                            href=output_band,
                            title=band,
                            media_type=MEDIA_TYPES_DICT[raster_format],
                        )
                    )

            # Remove the original raster file and its metadata
            remove(raster_path)
            remove_raster_metadata(dirname(raster_path))

            return asset_list


class ExtractedAssets(STACAssetGenerator):
    """
    Extracted STAC Asset Generator
    """

    type = "Extracted"

    def __init__(self) -> None:
        super().__init__()

    def extract_assets(self, obj_info: pd.DataFrame):
        """
        Get all the files with the same extension as the image file as assets
        """
        asset_list = []
        # File path
        raster_path = obj_info["image"].values[0]
        raster_dir = dirname(raster_path)
        # Get the files with the same extension as the image file
        files = [
            f for f in listdir(raster_dir) if f.endswith(raster_path.split(".")[-1])
        ]
        # Instantiate pystac asset and append it to the list
        for file in files:
            # Get the file extension
            raster_format = file.split(".")[-1]
            asset_list.append(
                pystac.Asset(
                    href=join(raster_dir, file),
                    title=basename(file),
                    media_type=MEDIA_TYPES_DICT[raster_format],
                )
            )

        return asset_list



"""
Module for raster STAC extensions object
"""

from typing import Union, Optional

import pystac
import rasterio
import pandas as pd

from pystac.extensions.raster import RasterExtension, RasterBand
from .base import STACExtensionObject


class RasterExtensionObject(STACExtensionObject):
    """
    Raster extension object
    """
    def __init__(self) -> None:
        super().__init__()

    def add_extension_to_object(
        self,
        obj: Union[pystac.Item, pystac.Asset],
        obj_info: Optional[pd.DataFrame] = None,
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        if not isinstance(obj, pystac.Asset):
            return obj
        else:
            raster_ext = RasterExtension.ext(obj, add_if_missing=True)
            src = rasterio.open(obj.href)
            bands = []
            for band in src.indexes:
                bands.append(
                    RasterBand.create(
                        nodata=src.nodatavals[band - 1],
                        data_type=src.dtypes[band - 1],
                        spatial_resolution=src.res,
                    )
                    if src.nodatavals
                    else RasterBand.create(
                        data_type=src.dtypes[band - 1], spatial_resolution=src.res
                    )
                )
            raster_ext.apply(bands=bands)

        return obj



import os
from pathlib import Path
from tqdm import tqdm

from ..auth import with_auth
from .retrieve import retrieve_dataset, retrieve_dataset_files
from ..repos import FilesAPIRepo, DatasetsAPIRepo
from ..curation.stac import STACDataFrame


@with_auth
def download_dataset(
    dataset_name,
    version=None,
    path=None,
    logger=print,
    assets=False,
    force=False,
    verbose=False,
    user=None,
    file=None,
):
    dataset = retrieve_dataset(dataset_name)
    if version is None:
        version = sorted(dataset["versions"], key=lambda v: v["version_id"])[-1][
            "version_id"
        ]
    else:
        assert version in [
            v["version_id"] for v in dataset["versions"]
        ], f"Version {version} not found"
    download_base_path = os.getenv(
        "EOTDL_DOWNLOAD_PATH", str(Path.home()) + "/.cache/eotdl/datasets"
    )
    if path is None:
        download_path = download_base_path + "/" + dataset_name + "/v" + str(version)
    else:
        download_path = path + "/" + dataset_name + "/v" + str(version)
    # check if dataset already exists
    if os.path.exists(download_path) and not force:
        os.makedirs(download_path, exist_ok=True)
        raise Exception(
            f"Dataset `{dataset['name']} v{str(version)}` already exists at {download_path}. To force download, use force=True or -f in the CLI."
        )
    if dataset["quality"] == 0:
        if file:
            raise NotImplementedError("Downloading a specific file is not implemented")
        dataset_files = retrieve_dataset_files(dataset["id"], version)
        repo = FilesAPIRepo()
        for file in tqdm(dataset_files, disable=verbose, unit="file", position=0):
            filename, file_version = file["filename"], file["version"]
            if verbose:
                logger(f"Downloading {file['filename']}...")
            dst_path = repo.download_file(
                dataset["id"],
                filename,
                user["id_token"],
                download_path,
                file_version,
                progress=True,
            )
            # if calculate_checksum(dst_path) != checksum:
            #     logger(f"Checksum for {file} does not match")
            if verbose:
                logger("Done")
    else:
        # raise NotImplementedError("Downloading a STAC dataset is not implemented")
        if verbose:
            logger("Downloading STAC metadata...")
        repo = DatasetsAPIRepo()
        gdf, error = repo.download_stac(
            dataset["id"],
            user["id_token"],
        )
        if error:
            raise Exception(error)
        df = STACDataFrame(gdf)
        # df.geometry = df.geometry.apply(lambda x: Polygon() if x is None else x)
        df.to_stac(download_path)
        # download assets
        if assets:
            if verbose:
                logger("Downloading assets...")
            repo = FilesAPIRepo()
            df = df.dropna(subset=["assets"])
            for row in tqdm(df.iterrows(), total=len(df)):
                for k, v in row[1]["assets"].items():
                    href = v["href"]
                    _, filename = href.split("/download/")
                    # will overwrite assets with same name :(
                    repo.download_file_url(
                        href, filename, f"{download_path}/assets", user["id_token"]
                    )
        else:
            if verbose:
                logger("To download assets, set assets=True or -a in the CLI.")
    return download_path


@with_auth
def download_file_url(url, path, progress=True, logger=print, user=None):
    repo = FilesAPIRepo()
    _, filename = url.split("/download/")
    return repo.download_file_url(
        url, filename, f"{path}/assets", user["id_token"], progress
    )



"""
Module for DEM STAC extensions object
"""

from .base import STACExtensionObject


class DEMExtensionObject(STACExtensionObject):
    """
    DEM STAC extension object
    """
    DEM_DATE_ACQUIRED = {
        "start_datetime": "2011-01-01T00:00:00Z",
        "end_datetime": "2015-01-07T00:00:00Z",
    }

    def __init__(self) -> None:
        super().__init__()



"""
Utils
"""

from sentinelhub import DataCollection, MosaickingOrder

from .evalscripts import EvalScripts


class SHParameters:
    """
    Sentinel Hub Parameters base class
    """
    def __init__(self):
        pass


class SHS2L2AParameters(SHParameters):
    """
    Sentinel-2-L2A parameters
    """
    DATA_COLLECTION = DataCollection.SENTINEL2_L2A
    RESOLUTION = 10
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L2A
    FIELDS = {
        "include": ["id", "properties.datetime", "properties.eo:cloud_cover"],
        "exclude": [],
    }
    FILTER = None


class SHS2L1CParameters(SHParameters):
    """
    Sentinel-2-L1C parameters
    """
    DATA_COLLECTION = DataCollection.SENTINEL2_L1C
    RESOLUTION = 10
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L1C
    FIELDS = {
        "include": ["id", "properties.datetime", "properties.eo:cloud_cover"],
        "exclude": [],
    }


class SHS1Parameters(SHParameters):
    """
    Sentinel-1 parameters
    """
    DATA_COLLECTION = DataCollection.SENTINEL1
    RESOLUTION = 3
    EVALSCRIPT = EvalScripts.SENTINEL_1
    FIELDS = (
        {
            "include": [
                "id",
                "properties.datetime",
                "sar:instrument_mode",
                "s1:polarization",
                "sat:orbit_state",
                "s1:resolution",
                "s1:timeliness",
            ],
            "exclude": [],
        },
    )
    FILTER = None


class SHDEMParameters(SHParameters):
    """
    Copernicus DEM parameters
    """
    DATA_COLLECTION = DataCollection.DEM_COPERNICUS_30
    RESOLUTION = 3
    MOSAICKING_ORDER = None
    EVALSCRIPT = EvalScripts.DEM
    FILTER = None
    FIELDS = None


SUPPORTED_SENSORS = ("sentinel-1-grd", "sentinel-2-l1c", "sentinel-2-l2a", "dem")

SH_PARAMETERS_DICT = {
    "sentinel-1-grd": SHS1Parameters,
    "sentinel-2-l1c": SHS2L1CParameters,
    "sentinel-2-l2a": SHS2L2AParameters,
    "dem": SHDEMParameters,
}



"""
Download imagery
"""

from datetime import datetime
from typing import Union, List

from .sentinelhub import (
    SHClient,
    SH_PARAMETERS_DICT,
    evaluate_sentinel_parameters,
    imagery_from_tmp_to_dir,
)
from .search import search_sentinel_imagery


def download_sentinel_imagery(
    output: str,
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    sensor: str,
) -> None:
    """
    Download Sentinel imagery
    """
    evaluate_sentinel_parameters(sensor, time_interval, bounding_box, output)

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    request = client.request_data(time_interval, bounding_box, parameters)
    client.download_data(request)
    imagery_from_tmp_to_dir(output)


def search_and_download_sentinel_imagery(
    output: str,
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    sensor: str,
) -> None:
    """
    Search and download Sentinel imagery
    """
    evaluate_sentinel_parameters(sensor, time_interval, bounding_box, output)

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    results = search_sentinel_imagery(time_interval, bounding_box, sensor)
    timestamps = [date.strftime("%Y-%m-%d") for date in results.get_timestamps()]

    requests_list = []
    for date in timestamps:
        requests_list.append(client.request_data(date, bounding_box, parameters))
    client.download_data(requests_list)
    imagery_from_tmp_to_dir(output)



"""
Module for SAR STAC extensions object
"""

from typing import Optional, Union

import pystac
import pandas as pd

from pystac.extensions.sar import SarExtension
from pystac.extensions.sar import FrequencyBand, Polarization

from .base import STACExtensionObject


class SarExtensionObject(STACExtensionObject):
    """
    SAR extension object
    """
    def __init__(self) -> None:
        super().__init__()
        self.polarizations = [Polarization.VV, Polarization.VH]
        self.polarizations_dict = {"VV": Polarization.VV, "VH": Polarization.VH}

    def add_extension_to_object(
        self,
        obj: Union[pystac.Item, pystac.Asset],
        obj_info: Optional[pd.DataFrame] = None,
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        # Add SAR extension to the item
        sar_ext = SarExtension.ext(obj, add_if_missing=True)
        if isinstance(obj, pystac.Item) or (
            isinstance(obj, pystac.Asset)
            and obj.title not in self.polarizations_dict.keys()
        ):
            polarizations = self.polarizations
        elif (
            isinstance(obj, pystac.Asset)
            and obj.title in self.polarizations_dict.keys()
        ):
            polarizations = [self.polarizations_dict[obj.title]]
        sar_ext.apply(
            instrument_mode="EW",
            polarizations=polarizations,
            frequency_band=FrequencyBand.C,
            product_type="GRD",
        )

        return obj



"""
Module for projection STAC extensions object
"""

from typing import Union

import pystac
import pandas as pd
import rasterio

from pystac.extensions.projection import ProjectionExtension
from .base import STACExtensionObject


class ProjExtensionObject(STACExtensionObject):
    """
    Projection extension object
    """
    def __init__(self) -> None:
        super().__init__()

    def add_extension_to_object(
        self, obj: Union[pystac.Item, pystac.Asset], obj_info: pd.DataFrame
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param obj_info: object info from the STACDataFrame
        """
        # Add raster extension to the item
        if isinstance(obj, pystac.Asset):
            return obj
        elif isinstance(obj, pystac.Item):
            proj_ext = ProjectionExtension.ext(obj, add_if_missing=True)
            ds = rasterio.open(obj_info["image"].values[0])
            # Assume all the bands have the same projection
            proj_ext.apply(
                epsg=ds.crs.to_epsg(),
                transform=ds.transform,
                shape=ds.shape,
            )

        return obj



import typer
from pathlib import Path

from ..datasets import (
    retrieve_datasets,
    ingest_dataset,
    download_dataset,
)

app = typer.Typer(help="Explore, ingest and download training datasets.")


@app.command()
def list(
    name: str = typer.Option(
        None, "--name", "-n", help="Filter the returned datasets by name"
    ),
    limit: int = typer.Option(
        None, "--limit", "-l", help="Limit the number of returned results"
    ),
):
    """
    Retrieve a list with all the datasets in the EOTDL.

    If using --name, it will filter the results by name. If no name is provided, it will return all the datasets.\n
    If using --limit, it will limit the number of results. If no limit is provided, it will return all the datasets.
    \n\n
    Examples\n
    --------\n
    $ eotdl datasets list\n
    $ eotdl datasets list --name YourModel --limit 5
    """
    try:
        datasets = retrieve_datasets(name, limit)
        typer.echo(datasets)
    except Exception as e:
        typer.echo(e)


@app.command()
def ingest(
    path: Path = typer.Option(
        ..., "--path", "-p", help="Path to the dataset to ingest"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the ingestion",
    ),
):
    """
    Ingest a dataset to the EOTDL.

    This command ingests the dataset to the EOTDL. The dataset must be a folder with the dataset files,
    and at least a metadata.yml file or a catalog.json file. If there are not these files, the ingestion
    will not work. All the files in the folder will be uploaded to the EOTDL.
    \n\n
    The following constraints apply to the dataset name:\n
    - It must be unique\n
    - It must be between 3 and 45 characters long\n
    - It can only contain alphanumeric characters and dashes.\n
    \n
    The metadata.yml file must contain the following fields:\n
    - name: the name of the dataset\n
    - authors: the author or authors of the dataset\n
    - license: the license of the dataset\n
    - source: the source of the dataset\n
    \n
    If using --verbose, it will print the progress of the ingestion.
    \n\n
    Examples\n
    --------\n
    $ eotdl dataset ingest --path /path/to/folder-with-dataset --verbose True
    """
    try:
        ingest_dataset(path, verbose, typer.echo)
    except Exception as e:
        typer.echo(e)


@app.command()
def get(
    dataset: str = typer.Argument(None, help="Name of the dataset to download"),
    path: str = typer.Option(
        None, "--path", "-p", help="Download the dataset to a specific output path"
    ),
    file: str = typer.Option(
        None, "--file", "-f", help="Download a specific file from the dataset"
    ),
    version: int = typer.Option(None, "--version", "-v", help="Dataset version"),
    assets: bool = typer.Option(
        False, "--assets", "-a", help="Download STAC assets from the dataset"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force download even if file exists"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the download",
    ),
):
    """
    Download a dataset from the EOTDL.
    \n\n
    If using --path, it will download the dataset to the specified path. If no path is provided, it will download to ~/.eotdl/datasets.\n
    If using --file, it will download the specified file. If no file is provided, it will download the entire dataset.\n
    If using --version, it will download the specified version. If no version is provided, it will download the latest version.\n
    If using --assets when the dataset is STAC, it will also download the STAC assets of the dataset. If not provided, it will only download the STAC metadata.\n
    If using --force, it will download the dataset even if the file already exists.\n
    If using --verbose, it will print the progress of the download.
    \n\n
    Examples\n
    --------\n
    $ eotdl dataset get YourDataset\n
    $ eotdl dataset get YourDataset --path /path/to/download --file dataset.zip --version 1 --assets True --force True --verbose True
    """
    try:
        dst_path = download_dataset(
            dataset, version, path, typer.echo, assets, force, verbose
        )
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


if __name__ == "__main__":
    app()



from ..repos import DatasetsAPIRepo, FilesAPIRepo


def retrieve_datasets(name=None, limit=None):
    api_repo = DatasetsAPIRepo()
    data, error = api_repo.retrieve_datasets(name, limit)
    if data and not error:
        datasets = [d["name"] for d in data] if data else []
        return datasets
    return []


def retrieve_dataset(name):
    repo = DatasetsAPIRepo()
    data, error = repo.retrieve_dataset(name)
    if error:
        raise Exception(error)
    return data


def retrieve_dataset_files(dataset_id, version):
    repo = FilesAPIRepo()
    data, error = repo.retrieve_files(dataset_id, "datasets", version)
    if error:
        raise Exception(error)
    return data


# def list_datasets(pattern=None):
#     datasets = retrieve_datasets()
#     if pattern:
#         regex = re.compile(rf".*{re.escape(pattern)}.*", re.IGNORECASE)
#         names = list(datasets.keys())
#         valid = [name for name in names if regex.search(name)]
#         return {name: datasets[name] for name in valid}
#     return datasets



"""
STAC module
"""

# from .stac import STACGenerator
# from .utils import format_time_acquired
# from .parsers import STACIdParser, StructuredParser, UnestructuredParser
from .dataframe import STACDataFrame, read_stac



from pydantic import BaseModel
from ....utils import calculate_checksum


class DownloadFile:
    def __init__(self, repo, retrieve_dataset, logger):
        self.repo = repo
        self.retrieve_dataset = retrieve_dataset
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        dataset: str
        file: str
        path: str = None
        user: dict
        checksum: str

    class Outputs(BaseModel):
        dst_path: str

    def __call__(self, inputs: Inputs) -> Outputs:
        dataset = self.retrieve_dataset(inputs.dataset)
        dst_path = self.repo.download_file(
            inputs.dataset, inputs.file, inputs.user["id_token"], inputs.path
        )
        checksum = calculate_checksum(dst_path)
        self.logger(f"Checksum: {checksum}")
        if dataset["checksum"] != checksum:
            self.logger("Checksums do not match")
        return self.Outputs(dst_path=dst_path)



"""
Search imagery
"""

from typing import Union, List
from datetime import datetime

from .sentinelhub import SHClient, SH_PARAMETERS_DICT, evaluate_sentinel_parameters


def search_sentinel_imagery(
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    sensor: str,
) -> None:
    """
    Search Sentinel imagery
    """
    evaluate_sentinel_parameters(
        sensor, time_interval, bounding_box, output_needed=False
    )

    client = SHClient()
    parameters = SH_PARAMETERS_DICT[sensor]()

    return client.search_data(bounding_box, time_interval, parameters)



"""
Module for managing the Sentinel Hub configuration and data access
"""

import json
from os import getenv
from os.path import exists
from sentinelhub import (
    SHConfig,
    SentinelHubCatalog,
    BBox,
    bbox_to_dimensions,
    CRS,
    SentinelHubRequest,
    SentinelHubDownloadClient,
    MimeType,
)

from ...repos.AuthRepo import AuthRepo
from .parameters import SHParameters
from ...tools.time_utils import prepare_time_interval
from ...tools.geo_utils import compute_image_size


class SHClient:
    """
    Client class to manage the Sentinel Hub Python interface.
    """

    def __init__(self) -> None:
        """ """
        self.config = SHConfig()
        if getenv("SH_CLIENT_ID") and getenv("SH_CLIENT_SECRET"):
            # If the user has provided the credentials in a .env file, we should save them
            self.config.sh_client_id = getenv("SH_CLIENT_ID")
            self.config.sh_client_secret = getenv("SH_CLIENT_SECRET")
        else:
            # If the user has not provided the credentials, we should check if
            # the user is logged in
            auth_repo = AuthRepo()
            creds_file = auth_repo.creds_path
            if not exists(creds_file):
                raise ValueError(
                    "Your are not logged in and have not provided Sentinel Hub credentials. Please, crete a .env file with your SH_CLIENT_ID and SH_CLIENT_SECRET, or login"
                )
            else:
                with open(creds_file, "r", encoding="utf-8") as f:
                    creds = json.load(f)
                if (
                    "SH_CLIENT_ID" not in creds.keys()
                    or "SH_CLIENT_SECRET" not in creds.keys()
                ):
                    raise ValueError(
                        "If you already had a Sentinel HUB account before accepting the Terms and Conditions, your SH credentials will NOT appear here. You can retrieve them from you Sentinel HUB dashboard)"
                    )
                else:
                    self.config.sh_client_id = creds["SH_CLIENT_ID"]
                    self.config.sh_client_secret = creds["SH_CLIENT_SECRET"]
        self.catalog = SentinelHubCatalog(config=self.config)
        self.tmp_dir = "/tmp/sentinelhub"

    def search_data(
        self, bounding_box: list, time_interval: list, parameters: SHParameters
    ) -> list:
        """
        Search data from Sentinel Hub
        """
        search_iterator = self.catalog.search(
            parameters.DATA_COLLECTION,
            bbox=BBox(bounding_box, crs=CRS.WGS84),
            time=time_interval,
            filter=parameters.FILTER,
            fields=parameters.FIELDS,
        )

        return search_iterator

    def request_data(
        self, time_interval, bounding_box: list, parameters: SHParameters
    ) -> list:
        """
        Request data from Sentinel Hub
        """
        time_interval = prepare_time_interval(time_interval)
        bounding_box, _ = compute_image_size(bounding_box, parameters)

        return SentinelHubRequest(
            data_folder=self.tmp_dir,
            evalscript=parameters.EVALSCRIPT,
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=parameters.DATA_COLLECTION,
                    time_interval=time_interval,
                    mosaicking_order=parameters.MOSAICKING_ORDER,
                )
            ],
            responses=[SentinelHubRequest.output_response("default", MimeType.TIFF)],
            bbox=bounding_box,
            size=bbox_to_dimensions(bounding_box, parameters.RESOLUTION),
            config=self.config,
        )

    def download_data(self, requests: list) -> list:
        """
        Download data from Sentinel Hub
        """
        download_client = SentinelHubDownloadClient(config=self.config)
        if not isinstance(requests, list):
            requests = [requests]
        download_requests = [request.download_list[0] for request in requests]
        data = download_client.download(download_requests)

        return data



"""
Utils for Sentinel Hub access
"""

import json

from os import makedirs
from datetime import datetime
from typing import Union, Optional
from glob import glob
from shutil import copyfile, rmtree

from .parameters import SUPPORTED_SENSORS
from ...tools.geo_utils import is_bounding_box, get_image_bbox
from ...tools.time_utils import is_time_interval, get_day_between


def evaluate_sentinel_parameters(
    sensor: str,
    time_interval: Union[str, datetime],
    bounding_box: list,
    output: Optional[str] = None,
    output_needed: Optional[bool] = True,
) -> None:
    """
    Evaluate parameters for Sentinel Hub access
    """
    if output_needed:
        if not output:
            raise ValueError("Output path must be specified.")
    if sensor not in SUPPORTED_SENSORS:
        raise ValueError(
            f"Sensor {sensor} is not supported. Supported sensors are: {SUPPORTED_SENSORS}"
        )
    if not time_interval:
        raise ValueError("Time interval must be specified.")
    else:
        if len(time_interval) == 2 and not is_time_interval(time_interval):
            raise ValueError(
                "Time interval must be a list or tuple with two elements in format YYYY-MM-DD."
            )
    if not bounding_box:
        raise ValueError("Bounding box must be specified.")
    else:
        if not is_bounding_box(bounding_box):
            raise ValueError(
                "Bounding box must be a list or tuple with four elements in format (lon_min, lat_min, lon_max, lat_max)."
            )


def imagery_from_tmp_to_dir(
    output_dir: str, tmp_dir: Optional[str] = "/tmp/sentinelhub"
) -> None:
    """
    Copy imagery from tmp to output dir
    """
    downloaded_files = glob(f"{tmp_dir}/**/response.tiff")
    assert len(downloaded_files) > 0, "No files downloaded"

    makedirs(output_dir, exist_ok=True)

    for downloaded_file in downloaded_files:
        request_json = downloaded_file.replace("response.tiff", "request.json")
        metadata = generate_raster_metadata(downloaded_file, request_json)
        if metadata["acquisition-date"]:
            output_filename = f"{metadata['type']}_{metadata['acquisition-date']}"
        else:
            output_filename = metadata["type"]

        copyfile(downloaded_file, f"{output_dir}/{output_filename}.tif")
        with open(f"{output_dir}/{output_filename}.json", "w", encoding="utf-8") as f:
            json.dump(metadata, f)

    rmtree(tmp_dir)


def generate_raster_metadata(raster: str, request_json: str) -> None:
    """
    Generate metadata for raster
    """
    bbox = get_image_bbox(raster)
    with open(request_json, "r", encoding="utf-8") as f:
        json_content = json.load(f)

    payload_data = json_content["request"]["payload"]["input"]["data"][0]
    sensor_type = payload_data["type"]
    # Get day between from and to
    if "timeRange" in payload_data["dataFilter"] and sensor_type != "dem":
        time_range = payload_data["dataFilter"]["timeRange"]
        acquisition_date = get_day_between(time_range["from"], time_range["to"])
    else:  # DEM data does not have a timeRange
        acquisition_date = None

    metadata = {
        "acquisition-date": acquisition_date,
        "bounding-box": bbox,
        "type": sensor_type,
    }

    return metadata



import typer
from .commands import auth, datasets, models
from . import __version__

app = typer.Typer(help="Welcome to EOTDL. Learn more at https://www.eotdl.com/")

app.add_typer(auth.app, name="auth")
app.add_typer(datasets.app, name="datasets")
app.add_typer(models.app, name="models")


@app.command()
def version():
    """
    Get EOTDL version.
    """
    typer.echo(f"EOTDL Version: {__version__}")


if __name__ == "__main__":
    app()



import requests
import os
from tqdm import tqdm
import hashlib

from ..repos import APIRepo


class FilesAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def ingest_files_batch(
        self,
        batch,  # ziped batch of files
        checksums,
        dataset_or_model_id,
        id_token,
        endpoint,
        version=None,
    ):
        url = self.url + f"{endpoint}/{dataset_or_model_id}/batch"
        if version is not None:
            url += "?version=" + str(version)
        reponse = requests.post(
            url,
            files={"batch": ("batch.zip", batch)},
            data={"checksums": checksums},
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(reponse)

    def add_files_batch_to_version(
        self,
        batch,
        dataset_or_model_id,
        version,
        id_token,
        endpoint,
    ):
        reponse = requests.post(
            self.url + f"{endpoint}/{dataset_or_model_id}/files?version={str(version)}",
            data={
                "filenames": [f["path"] for f in batch],
                "checksums": [f["checksum"] for f in batch],
            },
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(reponse)

    def ingest_file(
        self, file, dataset_or_model_id, id_token, checksum, endpoint, version=None
    ):
        url = self.url + f"{endpoint}/{dataset_or_model_id}"
        if version is not None:
            url += "?version=" + str(version)
        reponse = requests.post(
            url,
            files={"file": open(file, "rb")},
            data={"checksum": checksum},
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(reponse)

    def retrieve_files(self, dataset_or_model_id, endpoint, version=None):
        url = f"{self.url}{endpoint}/{dataset_or_model_id}/files"
        if version is not None:
            url += "?version=" + str(version)
        response = requests.get(url)
        return self.format_response(response)

    def download_file(
        self,
        dataset_or_model_id,
        file_name,
        id_token,
        path,
        file_version,
        endpoint="datasets",
        progress=False,
    ):
        url = self.url + f"{endpoint}/{dataset_or_model_id}/download/{file_name}"
        if file_version is not None:
            url += "?version=" + str(file_version)
        return self.download_file_url(url, file_name, path, id_token, progress=progress)

    def download_file_url(self, url, filename, path, id_token, progress=False):
        headers = {"Authorization": "Bearer " + id_token}
        path = f"{path}/{filename}"
        for i in range(1, len(path.split("/")) - 1):
            # print("/".join(path.split("/")[: i + 1]))
            os.makedirs("/".join(path.split("/")[: i + 1]), exist_ok=True)
        with requests.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024 * 1024 * 10
            progress = progress and total_size > 1024 * 1024 * 16
            if progress:
                progress_bar = tqdm(
                    total=total_size,
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                    position=1,
                )
            with open(path, "wb") as f:
                for chunk in r.iter_content(block_size):
                    if progress:
                        progress_bar.update(len(chunk))
                    if chunk:
                        f.write(chunk)
            if progress:
                progress_bar.close()
            return path

    # def ingest_file_url(self, file, dataset, id_token):
    #     reponse = requests.post(
    #         self.url + f"datasets/{dataset}/url",
    #         json={"url": file},
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if reponse.status_code != 200:
    #         return None, reponse.json()["detail"]
    #     return reponse.json(), None

    def prepare_large_upload(
        self, filename, dataset_or_model_id, checksum, id_token, endpoint
    ):
        response = requests.post(
            self.url + f"{endpoint}/{dataset_or_model_id}/uploadId",
            json={"filname": filename, "checksum": checksum},
            headers={"Authorization": "Bearer " + id_token},
        )
        if response.status_code != 200:
            raise Exception(response.json()["detail"])
        data = response.json()
        upload_id, parts = (
            data["upload_id"],
            data["parts"] if "parts" in data else [],
        )
        return upload_id, parts

    def get_chunk_size(self, content_size):
        # adapt chunk size to content size to avoid S3 limits (10000 parts, 500MB per part, 5TB per object)
        chunk_size = 1024 * 1024 * 10  # 10 MB (up to 100 GB, 10000 parts)
        if content_size >= 1024 * 1024 * 1024 * 100:  # 100 GB
            chunk_size = 1024 * 1024 * 100  # 100 MB (up to 1 TB, 10000 parts)
        elif content_size >= 1024 * 1024 * 1024 * 1000:  # 1 TB
            chunk_size = 1024 * 1024 * 500  # 0.5 GB (up to 5 TB, 10000 parts)
        return chunk_size

    def read_in_chunks(self, file_object, CHUNK_SIZE):
        while True:
            data = file_object.read(CHUNK_SIZE)
            if not data:
                break
            yield data

    def ingest_large_file(
        self, file_path, files_size, upload_id, id_token, parts, endpoint
    ):
        print(endpoint)
        # content_path = os.path.abspath(file)
        # content_size = os.stat(content_path).st_size
        chunk_size = self.get_chunk_size(files_size)
        total_chunks = files_size // chunk_size
        # upload chunks sequentially
        pbar = tqdm(
            self.read_in_chunks(open(file_path, "rb"), chunk_size),
            total=total_chunks,
        )
        index = 0
        for chunk in pbar:
            part = index // chunk_size + 1
            offset = index + len(chunk)
            index = offset
            if part not in parts:
                checksum = hashlib.md5(chunk).hexdigest()
                response = requests.post(
                    f"{self.url}{endpoint}/chunk/{upload_id}",
                    files={"file": chunk},
                    data={"part_number": part, "checksum": checksum},
                    headers={"Authorization": "Bearer " + id_token},
                )
                if response.status_code != 200:
                    raise Exception(response.json()["detail"])
            pbar.set_description(
                "{:.2f}/{:.2f} MB".format(
                    offset / 1024 / 1024, files_size / 1024 / 1024
                )
            )
        pbar.close()
        return

    def complete_upload(self, id_token, upload_id, version, endpoint):
        r = requests.post(
            f"{self.url}{endpoint}/complete/{upload_id}?version={version}",
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(r)

    # def delete_file(self, dataset_id, file_name, id_token):
    #     response = requests.delete(
    #         self.url + "datasets/" + dataset_id + "/file/" + file_name,
    #         headers={"Authorization": "Bearer " + id_token},
    #     )
    #     if response.status_code != 200:
    #         return None, response.json()["detail"]
    #     return response.json(), None



import os
from pathlib import Path
from tqdm import tqdm

from ..auth import with_auth
from .retrieve import retrieve_model, retrieve_model_files
from ..shared import calculate_checksum
from ..repos import FilesAPIRepo


@with_auth
def download_model(
    model_name,
    version=None,
    path=None,
    logger=None,
    assets=False,
    force=False,
    verbose=False,
    user=None,
    file=None,
):
    model = retrieve_model(model_name)
    if version is None:
        version = sorted(model["versions"], key=lambda v: v["version_id"])[-1][
            "version_id"
        ]
    else:
        assert version in [
            v["version_id"] for v in model["versions"]
        ], f"Version {version} not found"
    download_base_path = os.getenv(
        "EOTDL_DOWNLOAD_PATH", str(Path.home()) + "/.cache/eotdl/models"
    )
    if path is None:
        download_path = download_base_path + "/" + model_name + "/v" + str(version)
    else:
        download_path = path + "/" + model_name + "/v" + str(version)
    # check if model already exists
    if os.path.exists(download_path) and not force:
        os.makedirs(download_path, exist_ok=True)
        raise Exception(
            f"model `{model['name']} v{str(version)}` already exists at {download_path}. To force download, use force=True or -f in the CLI."
        )
    if model["quality"] == 0:
        if file:
            raise NotImplementedError("Downloading a specific file is not implemented")
            # files = [f for f in model["files"] if f["name"] == file]
            # if not files:
            #     raise Exception(f"File {file} not found")
            # if len(files) > 1:
            #     raise Exception(f"Multiple files with name {file} found")
            # dst_path = download(
            #     model,
            #     model["id"],
            #     file,
            #     files[0]["checksum"],
            #     download_path,
            #     user,
            # )
            # return Outputs(dst_path=dst_path)
        model_files = retrieve_model_files(model["id"], version)
        repo = FilesAPIRepo()
        for file in tqdm(model_files, disable=verbose, unit="file"):
            filename, file_version = file["filename"], file["version"]
            if verbose:
                logger(f"Downloading {file['filename']}...")
            dst_path = repo.download_file(
                model["id"],
                filename,
                user["id_token"],
                download_path,
                file_version,
                endpoint="models",
            )
            # if calculate_checksum(dst_path) != checksum:
            #     logger(f"Checksum for {file} does not match")
            if verbose:
                logger("Done")
        return "/".join(dst_path.split("/")[:-1])
    else:
        raise NotImplementedError("Downloading a STAC model is not implemented")
    #     logger("Downloading STAC metadata...")
    #     gdf, error = repo.download_stac(
    #         model["id"],
    #         user["id_token"],
    #     )
    #     if error:
    #         raise Exception(error)
    #     df = STACDataFrame(gdf)
    #     # df.geometry = df.geometry.apply(lambda x: Polygon() if x is None else x)
    #     path = path
    #     if path is None:
    #         path = download_base_path + "/" + model["name"]
    #     df.to_stac(path)
    #     # download assets
    #     if assets:
    #         logger("Downloading assets...")
    #         df = df.dropna(subset=["assets"])
    #         for row in tqdm(df.iterrows(), total=len(df)):
    #             id = row[1]["stac_id"]
    #             # print(row[1]["links"])
    #             for k, v in row[1]["assets"].items():
    #                 href = v["href"]
    #                 repo.download_file_url(
    #                     href, f"{path}/assets/{id}", user["id_token"]
    #                 )
    #     else:
    #         logger("To download assets, set assets=True or -a in the CLI.")
    #     return Outputs(dst_path=path)


# @with_auth
# def download_file_url(url, path, progress=True, logger=None, user=None):
#     api_repo = APIRepo()
#     download = DownloadFileURL(api_repo, logger, progress)
#     inputs = DownloadFileURL.Inputs(url=url, path=path, user=user)
#     outputs = download(inputs)
#     return outputs.dst_path



"""Implements the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`."""

from typing import Any, Dict, List, Optional, Generic, TypeVar, Union, Set
from shutil import rmtree
from os.path import dirname, exists

import traceback
import json
import random

import pystac

from tqdm import tqdm
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.label import LabelExtension
from pystac import STACValidationError
from pystac.cache import ResolvedObjectCache
from pystac.extensions.hooks import ExtensionHooks
from ....tools import make_links_relative_to_path

T = TypeVar("T", pystac.Item, pystac.Collection, pystac.Catalog)


SCHEMA_URI: str = "https://raw.githubusercontent.com/earthpulse/ml-dataset/main/json-schema/schema.json"
PREFIX: str = "ml-dataset:"


class MLDatasetExtension(
    pystac.Catalog,
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[
        Union[pystac.item.Item, pystac.collection.Collection, pystac.catalog.Catalog]
    ],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection`, :class:`~pystac.Item`, or :class:`~pystac.Catalog` with
    properties from the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`MLDatasetExtension`, use the
    :meth:`MLDatasetExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> ml_ext = MLDatasetExtension.ext(item)
    """

    catalog: pystac.Catalog
    """The :class:`~pystac.Catalog` being extended."""

    properties: Dict[str, Any]
    """The :class:`~pystac.Catalog` extra fields, including extension properties."""

    links: List[pystac.Link]
    """The list of :class:`~pystac.Link` objects associated with the
    :class:`~pystac.Catalog` being extended, including links added by this extension.
    """

    def __init__(self, catalog: pystac.Catalog):
        super().__init__(id=catalog.id, description=catalog.description)
        self._catalog = catalog
        self.id = catalog.id
        self.description = catalog.description
        self.title = catalog.title if catalog.title else None
        self.stac_extensions = (
            catalog.stac_extensions if catalog.stac_extensions else []
        )
        self.extra_fields = self.properties = (
            catalog.extra_fields if catalog.extra_fields else {}
        )
        self.links = catalog.links
        self._resolved_objects = ResolvedObjectCache()

    def apply(self, name: str = None) -> None:
        """
        Applies the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>` to the extended
        :class:`~pystac.Catalog`.
        """
        self.name = name

    @property
    def name(self) -> str:
        """
        Name of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}name"]

    @name.setter
    def name(self, v: str) -> None:
        """
        Set the name of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}name"] = v

    @property
    def tasks(self) -> List:
        """
        Tasks of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}tasks"]

    @tasks.setter
    def tasks(self, v: Union[list, tuple]) -> None:
        """
        Set the tasks of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}tasks"] = v

    @property
    def type(self) -> str:
        """
        Type of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}type"]

    @type.setter
    def type(self, v: str) -> None:
        """
        Set the type of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}type"] = v

    @property
    def inputs_type(self) -> str:
        """
        Inputs type of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}inputs-type"]

    @inputs_type.setter
    def inputs_type(self, v: str) -> None:
        """
        Set the inputs type of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}inputs-type"] = v

    @property
    def annotations_type(self) -> str:
        """
        Annotations type of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}annotations-type"]

    @annotations_type.setter
    def annotations_type(self, v: str) -> None:
        """
        Set the annotations type of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}annotations-type"] = v

    @property
    def splits(self) -> List[str]:
        """
        Splits of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}splits"]

    @splits.setter
    def splits(self, v: dict) -> None:
        """
        Set the splits of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}splits"] = v

    @property
    def quality_metrics(self) -> List[dict]:
        """
        Quality metrics of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}quality-metrics"]

    @quality_metrics.setter
    def quality_metrics(self, v: dict) -> None:
        """
        Set the quality metrics of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}quality-metrics"] = v

    @property
    def version(self) -> str:
        """
        Version of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}version"]

    @version.setter
    def version(self, v: str) -> None:
        """
        Set the version of the ML Dataset.
        """
        self.extra_fields[f"{PREFIX}version"] = v

    @classmethod
    def get_schema_uri(cls) -> str:
        """
        Get the JSON Schema URI that validates the extended object.
        """
        return SCHEMA_URI

    def add_metric(self, metric: dict) -> None:
        """Add a metric to this object's set of metrics.

        Args:
             metric : The metric to add.
        """
        if not self.extra_fields.get(f"{PREFIX}quality-metrics"):
            self.extra_fields[f"{PREFIX}quality-metrics"] = []

        if metric not in self.extra_fields[f"{PREFIX}quality-metrics"]:
            self.extra_fields[f"{PREFIX}quality-metrics"].append(metric)

    def add_metrics(self, metrics: List[dict]) -> None:
        """Add a list of metrics to this object's set of metrics.

        Args:
             metrics : The metrics to add.
        """
        for metric in metrics:
            self.add_metric(metric)

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False):
        """Extends the given STAC Object with properties from the
        :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.

        This extension can be applied to instances of :class:`~pystac.Catalog`,
        :class:`~pystac.Collection` or :class:`~pystac.Item`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.validate_has_extension(obj, add_if_missing)
            return CollectionMLDatasetExtension(obj)
        elif isinstance(obj, pystac.Catalog):
            cls.validate_has_extension(obj, add_if_missing)
            return MLDatasetExtension(obj)
        elif isinstance(obj, pystac.Item):
            cls.validate_has_extension(obj, add_if_missing)
            return ItemMLDatasetExtension(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))


class CollectionMLDatasetExtension(MLDatasetExtension[pystac.Collection]):
    """A concrete implementation of :class:`MLDatasetExtension` on an
    :class:`~pystac.Collection` that extends the properties of the Collection to include
    properties defined in the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.
    """

    collection: pystac.Collection
    properties: Dict[str, Any]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields
        self.properties[f"{PREFIX}split-items"] = []

    def __repr__(self) -> str:
        return f"<CollectionMLDatasetExtension Item id={self.collection.id}>"

    @property
    def splits(self) -> List[dict]:
        """
        Splits of the ML Dataset.
        """
        return self.extra_fields[f"{PREFIX}splits"]

    @splits.setter
    def splits(self, v: dict) -> None:
        """
        Set the splits of the ML Dataset.
        """
        self.properties[f"{PREFIX}split-items"] = v

    def add_split(self, v: dict) -> None:
        """
        Add a split to the ML Dataset.
        """
        self.properties[f"{PREFIX}split-items"].append(v)

    def create_and_add_split(
        self, split_data: List[pystac.Item], split_type: str
    ) -> None:
        """
        Create and add a split to the ML Dataset.
        """
        items_ids = [item.id for item in split_data]
        items_ids.sort()

        split = {"name": split_type, "items": items_ids}

        if not self.properties.get(f"{PREFIX}split-items"):
            self.properties[f"{PREFIX}split-items"] = []

        self.add_split(split)
        print(f"Generating {split_type} split...")
        for _item in tqdm(split_data):
            item = self.collection.get_item(_item.id)
            if item:
                item_ml = MLDatasetExtension.ext(item, add_if_missing=True)
                item_ml.split = split_type


class ItemMLDatasetExtension(MLDatasetExtension[pystac.Item]):
    """A concrete implementation of :class:`MLDatasetExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Machine Learning Dataset Extension <ml-dataset>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`MLDatasetExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: Dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    @property
    def split(self) -> str:
        """
        Split of the ML Dataset.
        """
        return self.properties[f"{PREFIX}split"]

    @split.setter
    def split(self, v: str) -> None:
        """
        Set the split of the ML Dataset.
        """
        self.properties[f"{PREFIX}split"] = v

    def __repr__(self) -> str:
        return f"<ItemMLDatasetExtension Item id={self.item.id}>"


class MLDatasetQualityMetrics:
    """
    ML Dataset Quality Metrics
    """

    @classmethod
    def calculate(cls, catalog: Union[pystac.Catalog, str]) -> None:
        """
        Calculate the quality metrics of the catalog
        """
        if isinstance(catalog, str):
            catalog = MLDatasetExtension(pystac.read_file(catalog))
        elif isinstance(catalog, pystac.Catalog):
            catalog = MLDatasetExtension(catalog)
        # Check the catalog has the extension
        if not MLDatasetExtension.has_extension(catalog):
            raise pystac.ExtensionNotImplemented(
                f"MLDatasetExtension does not apply to type '{type(catalog).__name__}'"
            )

        try:
            catalog.add_metric(cls._search_spatial_duplicates(catalog))
            catalog.add_metric(cls._get_classes_balance(catalog))
        except AttributeError as exc:
            raise pystac.ExtensionNotImplemented(
                f"The catalog does not have the required properties or the ML-Dataset extension to calculate the metrics: {exc}"
            )
        finally:
            catalog.make_all_asset_hrefs_relative()

        try:
            print("Validating and saving...")
            catalog.validate()
            destination = dirname(catalog.get_self_href())
            rmtree(
                destination
            )  # Remove the old catalog and replace it with the new one
            catalog.set_root(catalog)
            catalog.normalize_and_save(root_href=destination)
            print("Success!")
        except STACValidationError:
            # Return full callback
            traceback.print_exc()

    @staticmethod
    def _search_spatial_duplicates(catalog: pystac.Catalog):
        """
        Search for spatial duplicates in the catalog
        """
        items = list(
            set(
                [
                    item
                    for item in tqdm(
                        catalog.get_items(recursive=True),
                        desc="Looking for spatial duplicates...",
                    )
                    if not LabelExtension.has_extension(item)
                ]
            )
        )

        # Initialize the spatial duplicates dict
        spatial_duplicates = {"name": "spatial-duplicates", "values": [], "total": 0}

        items_bboxes = {}
        for item in items:
            # Get the item bounding box
            bbox = str(item.bbox)
            # If the bounding box is not in the items dict, add it
            if bbox not in items_bboxes.keys():
                items_bboxes[bbox] = item.id
            # If the bounding box is already in the items dict, add it to the duplicates dict
            else:
                spatial_duplicates["values"].append(
                    {"item": item.id, "duplicate": items_bboxes[bbox]}
                )
                spatial_duplicates["total"] += 1

        return spatial_duplicates

    @staticmethod
    def _get_classes_balance(catalog: pystac.Catalog) -> dict:
        """
        Get the classes balance of the catalog
        """

        def get_label_properties(items: List[pystac.Item]) -> List:
            """
            Get the label properties of the catalog
            """
            label_properties = []
            for label in items:
                label_ext = LabelExtension.ext(label)
                for prop in label_ext.label_properties:
                    if prop not in label_properties:
                        label_properties.append(prop)

            return label_properties

        catalog.make_all_asset_hrefs_absolute()

        labels = list(
            set(
                [
                    item
                    for item in tqdm(
                        catalog.get_items(recursive=True),
                        desc="Calculating classes balance...",
                    )
                    if LabelExtension.has_extension(item)
                ]
            )
        )

        # Initialize the classes balance dict
        classes_balance = {"name": "classes-balance", "values": []}
        label_properties = get_label_properties(labels)

        for prop in label_properties:
            property_balance = {"name": prop, "values": []}
            properties = {}
            for label in labels:
                if 'labels' not in label.assets:
                    continue
                asset_path = label.assets["labels"].href
                # Open the linked geoJSON to obtain the label properties
                try:
                    with open(asset_path, mode="r", encoding="utf-8") as f:
                        label_data = json.load(f)
                except FileNotFoundError:
                    raise FileNotFoundError(
                        f"The file {asset_path} does not exist. Make sure the assets hrefs are correct"
                    )
                # Get the property
                for feature in label_data["features"]:
                    if prop in feature["properties"]:
                        property_value = feature["properties"][prop]
                    else:
                        if feature["properties"]["labels"]:
                            property_value = feature["properties"]["labels"][0]
                        else:
                            continue
                    if property_value not in properties:
                        properties[property_value] = 0
                    properties[property_value] += 1

            # Create the property balance dict
            total_labels = sum(properties.values())
            for key, value in properties.items():
                property_balance["values"].append(
                    {
                        "class": key,
                        "total": value,
                        "percentage": int(value / total_labels * 100),
                    }
                )

            classes_balance["values"].append(property_balance)

        catalog.make_all_asset_hrefs_relative()

        return classes_balance


class MLDatasetExtensionHooks(ExtensionHooks):
    """
    ML Dataset Extension Hooks
    """
    schema_uri: str = SCHEMA_URI
    prev_extension_ids: Set[str] = set()
    stac_object_types = {
        pystac.STACObjectType.CATALOG,
        pystac.STACObjectType.COLLECTION,
        pystac.STACObjectType.ITEM,
    }


STORAGE_EXTENSION_HOOKS: ExtensionHooks = MLDatasetExtensionHooks()


def add_ml_extension(
    catalog: Union[pystac.Catalog, str],
    destination: Optional[str] = None,
    splits: Optional[bool] = False,
    splits_collection_id: Optional[str] = "labels",
    splits_names: Optional[list] = ("Training", "Validation", "Test"),
    split_proportions: Optional[List[int]] = (80, 10, 10),
    catalog_type: Optional[pystac.CatalogType] = pystac.CatalogType.SELF_CONTAINED,
    **kwargs,
) -> None:
    """
    Adds the ML Dataset extension to a STAC catalog.
    """
    if not isinstance(catalog, pystac.Catalog) and isinstance(catalog, str):
        catalog = pystac.read_file(catalog)
    elif isinstance(catalog, pystac.Catalog):
        pass
    else:
        raise pystac.ExtensionTypeError(
            f"MLDatasetExtension does not apply to type '{type(catalog).__name__}'"
        )

    catalog_ml_dataset = MLDatasetExtension.ext(catalog, add_if_missing=True)
    if destination:
        catalog_ml_dataset.set_self_href(destination + "/catalog.json")
    else:
        destination = dirname(catalog.get_self_href())
    catalog_ml_dataset.set_root(catalog_ml_dataset)

    # Set extension properties
    for key, value in kwargs.items():
        setattr(catalog_ml_dataset, key, value)

    # Make splits if needed
    if splits:
        catalog_ml_dataset.splits = splits_names  # Add the splits names to the catalog
        train_size, test_size, val_size = split_proportions
        splits_collection = catalog.get_child(
            splits_collection_id
        )  # Get the collection to split
        if not splits_collection:
            raise AttributeError(
                f"The catalog does not have a collection with the id {splits_collection_id}"
            )
        make_splits(
            splits_collection,
            train_size=train_size,
            test_size=test_size,
            val_size=val_size,
            **kwargs,
        )

    # Normalize the ref on the same folder
    if destination:
        catalog_ml_dataset = make_links_relative_to_path(
            destination, catalog_ml_dataset
        )

    try:
        print("Validating and saving...")
        catalog_ml_dataset.validate()
        rmtree(destination) if exists(
            destination
        ) else None  # Remove the old catalog and replace it with the new one
        catalog_ml_dataset.normalize_and_save(
            root_href=destination, catalog_type=catalog_type
        )
        print("Success!")
    except STACValidationError:
        # Return full callback
        traceback.print_exc()


def make_splits(
    labels_collection: Union[CollectionMLDatasetExtension, pystac.Collection, str],
    splits_names: Optional[List[str]] = ("Training", "Validation", "Test"),
    splits_proportions: Optional[List[int]] = (80, 10, 10),
    verbose: Optional[bool] = True,
    **kwargs: Optional[dict],
) -> None:
    """
    Makes the splits of the labels collection.
    """
    if isinstance(labels_collection, str):
        labels_collection = pystac.read_file(labels_collection)

    train_size, test_size, val_size = splits_proportions

    if train_size + test_size + val_size != 100:
        raise ValueError("The sum of the splits must be 100")

    # Get all items in the labels collection
    items = list(labels_collection.get_items(recursive=True))

    # Calculate indices to split the items
    length = len(items)
    idx_train = int(train_size / 100 * length)
    idx_test = int(test_size / 100 * length)
    if val_size:
        idx_val = int(val_size / 100 * length)

    print("Generating splits...")
    if verbose:
        print(f"Total size: {length}")
        print(f"Train size: {idx_train}")
        print(f"Test size: {idx_test}")
        if val_size:
            print(f"Validation size: {idx_val}")

    # Make sure the items are shuffled
    random.shuffle(items)

    # Split the items
    train_items = items[:idx_train]
    test_items = items[idx_train: idx_train + idx_test]
    if val_size:
        val_items = items[idx_train + idx_test: idx_train + idx_test + idx_val]

    # Create the splits in the collection
    labels_collection = MLDatasetExtension.ext(labels_collection, add_if_missing=True)
    for split_type, split_data in zip(
        splits_names, [train_items, test_items, val_items]
    ):
        labels_collection.create_and_add_split(split_data, split_type)

    print("Success on splits generation!")



import requests

from ..repos import APIRepo


class ModelsAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def retrieve_models(self, name, limit):
        url = self.url + "models"
        if name is not None:
            url += "?match=" + name
        if limit is not None:
            if name is None:
                url += "?limit=" + str(limit)
            else:
                url += "&limit=" + str(limit)
        response = requests.get(url)
        return self.format_response(response)

    def create_model(self, metadata, id_token):
        response = requests.post(
            self.url + "models",
            json=metadata,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)

    def retrieve_model(self, name):
        response = requests.get(self.url + "models?name=" + name)
        return self.format_response(response)

    def create_version(self, model_id, id_token):
        response = requests.post(
            self.url + "models/version/" + model_id,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)



"""
STAC extensions module
"""

from .sar import SarExtensionObject
from .raster import RasterExtensionObject
from .projection import ProjExtensionObject
from .dem import DEMExtensionObject
from .eo import EOS2ExtensionObject
from .label import LabelExtensionObject, ImageNameLabeler, ScaneoLabeler
from .ml_dataset import add_ml_extension, MLDatasetQualityMetrics


SUPPORTED_EXTENSIONS = ("eo", "sar", "proj", "raster")


type_stac_extensions_dict = {
    "sar": SarExtensionObject(),
    "eo": EOS2ExtensionObject(),
    "dem": DEMExtensionObject(),
    "raster": RasterExtensionObject(),
    "proj": ProjExtensionObject(),
}



"""
Geo Utils
"""

import tarfile
from typing import Union
from statistics import mean

import geopandas as gpd
import rasterio
import rasterio.warp

from shapely import geometry
from shapely.geometry import box, Polygon, shape
from pyproj import Transformer
from sentinelhub import BBox, CRS, bbox_to_dimensions
from pandas import isna


def is_bounding_box(bbox: list) -> bool:
    """
    Check if the given bounding box is a bounding box and is valid
    """
    if not isinstance(bbox, (list, tuple)) or len(bbox) != 4:
        return False

    for value in bbox:
        if not isinstance(value, (int, float)):
            return False

    minx, miny, maxx, maxy = bbox
    if minx >= maxx or miny >= maxy:
        return False

    return True


def compute_image_size(bounding_box, parameters):
    """
    Compute the image size from the bounding box and the resolution
    """
    bbox = BBox(bbox=bounding_box, crs=CRS.WGS84)
    bbox_size = bbox_to_dimensions(bbox, resolution=parameters.RESOLUTION)

    return bbox, bbox_size


def get_image_bbox(raster: Union[tarfile.ExFileObject, str]):
    """
    Get the bounding box of a raster
    """
    with rasterio.open(raster) as src:
        bounds = src.bounds
        dst_crs = "EPSG:4326"
        left, bottom, right, top = rasterio.warp.transform_bounds(
            src.crs, dst_crs, *bounds
        )
        bbox = [left, bottom, right, top]
    return bbox


def get_image_resolution(raster: Union[tarfile.ExFileObject, str]):
    """
    Get the resolution of a raster
    """
    with rasterio.open(raster) as src:
        resolution = src.res
    return resolution


def bbox_to_coordinates(bounding_box: list) -> list:
    """
    Convert a bounding box to a list of polygon coordinates
    """
    polygon_coordinates = [
        (bounding_box[0], bounding_box[1]),  # bottom left
        (bounding_box[0], bounding_box[3]),  # top left
        (bounding_box[2], bounding_box[3]),  # top right
        (bounding_box[2], bounding_box[1]),  # bottom right
        (bounding_box[0], bounding_box[1]),  # back to bottom left
    ]

    return polygon_coordinates


def bbox_to_polygon(bounding_box: list) -> Polygon:
    """
    Convert a bounding box to a shapely polygon
    """
    polygon = box(bounding_box[0], bounding_box[1], bounding_box[2], bounding_box[3])

    return polygon


from_4326_transformer = Transformer.from_crs("EPSG:4326", "EPSG:3857")
from_3857_transformer = Transformer.from_crs("EPSG:3857", "EPSG:4326")


def bbox_from_centroid(
    x: Union[int, float],
    y: Union[int, float],
    pixel_size: Union[int, float],
    width: Union[int, float],
    height: Union[int, float],
) -> list:
    """
    Generate a bounding box from a centroid, pixel size and image dimensions.

    Params
    ------
    x: int or float
        x coordinate of the centroid
    y: int or float
        y coordinate of the centroid
    pixel_size: int or float
        pixel size in meters
    width: int or float
        width of the image in pixels
    height: int or float
        height of the image in pixels

    Returns
    -------
    bounding_box: list
        list with the bounding box coordinates
    """
    width_m = width * pixel_size
    heigth_m = height * pixel_size

    # Transform the centroid coordinates to meters
    centroid_m = from_4326_transformer.transform(x, y)

    # Calculate the bounding box coordinates
    min_x = centroid_m[0] - width_m / 2
    min_y = centroid_m[1] - heigth_m / 2
    max_x = centroid_m[0] + width_m / 2
    max_y = centroid_m[1] + heigth_m / 2

    # Convert the bounding box coordinates back to degrees
    min_x, min_y = from_3857_transformer.transform(min_x, min_y)
    max_x, max_y = from_3857_transformer.transform(max_x, max_y)

    return [min_y, min_x, max_y, max_x]


def generate_bounding_box(geom: geometry.point.Point, differences: list) -> list:
    """
    Generate the bounding box of a given point using the difference
    between the maximum and mininum coordinates of the bounding box

    :param geom: shapely geometry object of the point which we want to
                generate the bounding box.
    :param differences: list with the difference between the maximum
                and minimum longitude and latitude coordinates.
    :return: list with the resulting bounding box from the computing.
    """
    long_diff, lat_diff = differences[0], differences[1]
    lon, lat = geom.x, geom.y

    bbox = (
        lon - (long_diff / 2),
        lat - (lat_diff / 2),
        lon + (long_diff / 2),
        lat + (lat_diff / 2),
    )

    # Round the coordinates to 6 decimals
    bounding_box = [round(i, 6) for i in bbox]

    return bounding_box


def calculate_average_coordinates_distance(bounding_box_by_location: dict) -> list:
    """
    Calculate the mean distance between maximum and minixum longitude
    and latitude of the bounding boxes from the existing locations.
    This is intended to use these mean distance to generate the bounding
    boxes of the new locations given a centroid.

    :param bounding_box_by_location: dictionary with format
    location_id : bounding_box for the existing locations in
    the sen12floods dataset.
    :return mean_long_diff, mean_lat_diff: mean longitude
    and latitude difference in the bounding boxes
    """
    long_diff_list, lat_diff_list = [], []

    for bbox in bounding_box_by_location.values():
        long_diff = bbox[2] - bbox[0]
        long_diff_list.append(long_diff)
        lat_diff = bbox[3] - bbox[1]
        lat_diff_list.append(lat_diff)

    mean_long_diff = mean(long_diff_list)
    mean_lat_diff = mean(lat_diff_list)

    return mean_long_diff, mean_lat_diff


def generate_new_locations_bounding_boxes(
    gdf: gpd.GeoDataFrame, mean_differences: list, latest_id: int
) -> dict:
    """
    Generate the bounding box of every new location, using
    the mean difference between the maximum and minimum calculated
    longitude and latitude. This function also returns the time
    interval which we want to request from Sentinel Hub Services.

    :param gdf: GeoDataFrame wiht the new locations that
    are going to be added to the dataset
    :param mean_differences: list with the longitude
    and latitude mean differences, which are going to be used to generate
    the bounding boxes.
    :return: bbox_by_new_location: dict with format {<location_id>:
    {'bounding_box': list(), 'time_interval': list()}, ... }
    that contains the bounding box and time interval of the imagery for each location
    """
    bbox_by_new_location = {}

    for _, row in gdf.iterrows():
        new_location_id = str(latest_id + 1)
        time_interval = row["Began"].strftime("%Y-%m-%d"), row["Ended"].strftime(
            "%Y-%m-%d"
        )
        bbox = generate_bounding_box(row["geometry"], mean_differences)
        bbox_by_new_location[new_location_id] = {
            "bounding_box": bbox,
            "time_interval": time_interval,
        }
        latest_id += 1

    return bbox_by_new_location


def convert_df_geom_to_shape(row):
    """
    Convert the geometry of a dataframe row to a shapely shape
    """
    if not isna(row["geometry"]):
        geo = shape(row["geometry"])
        wkt = geo.wkt
    else:
        wkt = "POLYGON EMPTY"

    return wkt



"""
Module for the STAC dataframe
"""

import pandas as pd
import geopandas as gpd
import pystac
import json
import os
from xcube_geodb.core.geodb import GeoDBClient
from geomet import wkt
from os.path import join
from os import makedirs

from math import isnan
from .utils import convert_df_geom_to_shape, get_all_children


class STACDataFrame(gpd.GeoDataFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_stac_file(self, stac_file):
        """
        Create a STACDataFrame from a STAC file
        """
        return read_stac(stac_file)

    @classmethod
    def from_geodb(
        self,
        server_url: str,
        server_port: int | str,
        client_id: str,
        client_secret: str,
        auth_aud: str,
        collection: str,
        database: str = None,
    ):
        """
        Create a STACDataFrame from a GeoDB collection

        :param server_url: GeoDB server url
        :param server_port: GeoDB server port
        :param client_id: GeoDB client id
        :param client_secret: GeoDB client secret
        :param auth_aud: GeoDB auth aud
        :param collection: GeoDB collection
        :param database: GeoDB database
        """
        geodb_client = GeoDBClient(
            server_url=server_url,
            server_port=server_port,
            client_id=client_id,
            client_secret=client_secret,
            auth_aud=auth_aud,
        )

        data = geodb_client.get_collection(collection, database=database)

        return STACDataFrame(data, crs="EPSG:4326")

    def ingest(
        self,
        collection: str,
        server_url: str = os.environ["SERVER_URL"],
        server_port: int = os.environ["SERVER_PORT"],
        client_id: str = os.environ["CLIENT_ID"],
        client_secret: str = os.environ["CLIENT_SECRET"],
        auth_aud: str = os.environ["AUTH_DOMAIN"],
        database: str = None,
    ):
        """
        Create a GeoDB collection from a STACDataFrame

        :param collection: dataset name (GeoDB collection)
        :param server_url: GeoDB server url
        :param server_port: GeoDB server port
        :param client_id: GeoDB client id
        :param client_secret: GeoDB client secret
        :param auth_aud: GeoDB auth aud
        :param database: GeoDB database
        """

        geodb_client = GeoDBClient(
            server_url=server_url,
            server_port=server_port,
            client_id=client_id,
            client_secret=client_secret,
            auth_aud=auth_aud,
        )

        # TODO: check name is unique (use eotdl-cli)

        # TODO: ingest assets (only if local)
        # TODO: rename assets in the dataframe with URLs (only if local)

        # ingest to geodb

        # Check if the collection already exists
        if geodb_client.collection_exists(collection, database=database):
            # geodb_client.drop_collection(collection, database=database)
            raise Exception(f"Collection {collection} already exists")

        # Rename the column id to stac_id, to avoid conflicts with the id column
        self.rename(columns={"id": "stac_id"}, inplace=True)
        # Fill the NaN with '' to avoid errors, except in the geometry column
        copy = self.copy()
        columns_to_fill = copy.columns.drop("geometry")
        self[columns_to_fill] = self[columns_to_fill].fillna("")

        # Create the collection if it does not exist
        # and insert the data
        collections = {collection: self._create_collection_structure(self.columns)}
        geodb_client.create_collections(collections, database=database)

        geodb_client.insert_into_collection(collection, database=database, values=self)

        # TODO: save data in eotdl

    def _create_collection_structure(self, columns: list) -> dict:
        """
        Create the schema structure of a GeoDB collection from a STACDataFrame

        :param columns: columns of the STACDataFrame
        """
        stac_collection = {"crs": 4326, "properties": {}}

        for column in columns:
            if column not in ("geometry", "id"):
                stac_collection["properties"][column] = "json"

        return stac_collection

    def to_stac(self):
        """
        Create a STAC catalog and children from a STACDataFrame
        """
        df = self.copy()

        if "id" in df.columns and "stac_id" in df.columns:
            id_column = "stac_id"
            stac_id_exists = True
        else:
            id_column = "id"
            stac_id_exists = False

        # First, create the catalog and its folder, if exists
        catalog_df = df[df["type"] == "Catalog"]

        if catalog_df.empty:
            root_output_folder = "output"
            makedirs(root_output_folder, exist_ok=True)
        else:
            for index, row in catalog_df.iterrows():
                root_output_folder = row[id_column]
                makedirs(root_output_folder, exist_ok=True)
                row_json = row.to_dict()

                # Curate the json row
                row_json = self.curate_json_row(row_json, stac_id_exists)

                with open(join(root_output_folder, "catalog.json"), "w") as f:
                    json.dump(row_json, f)

        # Second, create the collections and their folders, if exist
        collections = dict()
        collections_df = df[df["type"] == "Collection"]
        for index, row in collections_df.iterrows():
            stac_output_folder = join(root_output_folder, row[id_column])
            collections[row[id_column]] = stac_output_folder
            makedirs(stac_output_folder, exist_ok=True)
            row_json = row.to_dict()

            # Curate the json row
            row_json = self.curate_json_row(row_json, stac_id_exists)

            with open(join(stac_output_folder, "collection.json"), "w") as f:
                json.dump(row_json, f)

        # Then, create the items and their folders, if exist
        features_df = df[df["type"] == "Feature"]
        for index, row in features_df.iterrows():
            collection = row["collection"]
            stac_output_folder = join(collections[collection], row[id_column])

            # Convert the geometry from WKT back to geojson
            row["geometry"] = row["geometry"].wkt
            row["geometry"] = wkt.loads(row["geometry"])
            makedirs(stac_output_folder, exist_ok=True)
            row_json = row.to_dict()

            # Curate the json row
            row_json = self.curate_json_row(row_json, stac_id_exists)

            with open(join(stac_output_folder, f'{row_json["id"]}.json'), "w") as f:
                json.dump(row_json, f)

    def curate_json_row(self, row: dict, stac_id_exists: bool) -> dict:
        """
        Curate the json row of a STACDataFrame, in order to generate a valid STAC file

        :param row: row of a STACDataFrame
        :param stac_id_exists: if the stac_id column exists
        """
        keys_to_remove = list()

        # Remove the created_at and modified_at columns, if the STACDataFrame comes from GeoDB
        for i in "created_at", "modified_at":
            if i in row.keys():
                keys_to_remove.append(i)

        # Rename the stac_id column to id, to avoid conflicts with the id column
        if stac_id_exists:
            row["id"] = row["stac_id"]
            del row["stac_id"]

        # Remove the NaN values and empty strings
        for k, v in row.items():
            if (isinstance(v, float) and isnan(v)) or v == "":
                keys_to_remove.append(k)
        for key in keys_to_remove:
            del row[key]
        del row["geometry"]

        return row


def read_stac(
    stac_file: pystac.Catalog | pystac.Collection | str,
    geometry_column: str = "geometry",
) -> STACDataFrame:
    """
    Read a STAC file and return a STACDataFrame

    :param stac_file: STAC file to read
    :param geometry_column: name of the geometry column
    """
    if isinstance(stac_file, str):
        stac_file = pystac.read_file(stac_file)
    children = get_all_children(stac_file)

    # Convert Dataframe to STACDataFrame
    dataframe = pd.DataFrame(children)
    dataframe[geometry_column] = dataframe.apply(convert_df_geom_to_shape, axis=1)
    stac_dataframe = STACDataFrame(
        dataframe,
        crs="EPSG:4326",
        geometry=gpd.GeoSeries.from_wkt(dataframe[geometry_column]),
    )

    return stac_dataframe



from pathlib import Path
import os
from tqdm import tqdm
import zipfile
import io
from glob import glob
import os

from ..repos import FilesAPIRepo
from ..shared import calculate_checksum


def retrieve_files(folder):
    # get all files in directory recursively
    items = [Path(item) for item in glob(str(folder) + "/**/*", recursive=True)]
    if not any(item.name == "metadata.yml" for item in items):
        raise Exception("metadata.yml not found in directory")
    # remove directories
    items = [item for item in items if not item.is_dir()]
    if len(items) == 0:
        raise Exception("No files found in directory")
    return items


def prepare_item(item, folder):
    return {
        "filename": item.name,
        "path": str(item.relative_to(folder)),
        "absolute_path": item.absolute(),
        "size": os.path.getsize(item.absolute()),
        "checksum": calculate_checksum(item.absolute()),
    }


def generate_batches(files, max_batch_size=1024 * 1024 * 10, max_batch_files=10):
    batches = []
    for item in tqdm(files):
        if not batches:
            batches.append([item])
            continue
        if max_batch_size:
            size_check = sum([i["size"] for i in batches[-1]]) < max_batch_size
        else:
            size_check = True
        if size_check and len(batches[-1]) < max_batch_files:
            batches[-1].append(item)
        else:
            batches.append([item])
    return batches


def compress_batch(batch):
    memory_file = io.BytesIO()
    with zipfile.ZipFile(memory_file, "w") as zf:
        for f in batch:
            zf.write(f["absolute_path"], arcname=f["path"])
    memory_file.seek(0)
    return memory_file


def generate_files_lists(
    items, folder, dataset_or_model_id, endpoint, logger, max_size=1024 * 1024 * 16
):
    files_repo = FilesAPIRepo()
    current_files, error = files_repo.retrieve_files(dataset_or_model_id, endpoint)
    # print(len(current_files), len(items) - len(current_files))
    # print(current_files, error)
    if error:
        current_files = []
    # generate list of files to upload
    logger("generating list of files to upload...")
    upload_files, existing_files, large_files = [], [], []
    current_names = [f["filename"] for f in current_files]
    current_checksums = [f["checksum"] for f in current_files]
    for item in tqdm(items):
        data = prepare_item(item, folder)
        if data["path"] in current_names and data["checksum"] in current_checksums:
            existing_files.append(data)
        else:
            if data["size"] > max_size:
                large_files.append(data)
            else:
                upload_files.append(data)
    # TODO: should ingest new version if files removed
    if len(upload_files) == 0 and len(large_files) == 0:
        raise Exception("No new files to upload")
    return upload_files, existing_files, large_files


def create_new_version(repo, dataset_or_model_id, user):
    data, error = repo.create_version(dataset_or_model_id, user["id_token"])
    if error:
        raise Exception(error)
    return data["version"]


def ingest_files(repo, dataset_or_model_id, folder, verbose, logger, user, endpoint):
    files_repo = FilesAPIRepo()
    logger(f"Uploading directory {folder}...")
    items = retrieve_files(folder)
    # retrieve files
    upload_files, existing_files, large_files = generate_files_lists(
        items, folder, dataset_or_model_id, endpoint, logger
    )
    logger(f"{len(upload_files) + len(large_files)} new files will be ingested")
    logger(f"{len(existing_files)} files already exist in dataset")
    logger(f"{len(large_files)} large files will be ingested separately")
    # create new version
    version = create_new_version(repo, dataset_or_model_id, user)
    logger("New version created, version: " + str(version))
    # ingest new large files
    if len(large_files) > 0:
        logger("ingesting large files...")
        for file in large_files:
            logger("ingesting file: " + file["path"])
            upload_id, parts = files_repo.prepare_large_upload(
                file["path"],
                dataset_or_model_id,
                file["checksum"],
                user["id_token"],
                endpoint,
            )
            # print(upload_id, parts)
            files_repo.ingest_large_file(
                file["absolute_path"],
                file["size"],
                upload_id,
                user["id_token"],
                parts,
                endpoint,
            )
            files_repo.complete_upload(user["id_token"], upload_id, version, endpoint)
    # ingest new small files in batches
    if len(upload_files) > 0:
        logger("generating batches...")
        batches = generate_batches(upload_files)
        logger(
            f"Uploading {len(upload_files)} small files in {len(batches)} batches..."
        )
        repo = FilesAPIRepo()
        for batch in tqdm(
            batches, desc="Uploading batches", unit="batches", disable=verbose
        ):
            # compress batch
            memory_file = compress_batch(batch)
            # ingest batch
            data, error = repo.ingest_files_batch(
                memory_file,
                [f["checksum"] for f in batch],
                dataset_or_model_id,
                user["id_token"],
                endpoint,
                version,
            )
    # ingest existing files
    if len(existing_files) > 0:
        batches = generate_batches(existing_files, max_batch_size=None)
        for batch in tqdm(
            batches,
            desc="Ingesting existing files",
            unit="batches",
            disable=verbose,
        ):
            data, error = files_repo.add_files_batch_to_version(
                batch,
                dataset_or_model_id,
                version,
                user["id_token"],
                endpoint,
            )
            if error:
                raise Exception(error)
    return data



from pydantic import BaseModel
import json
from pathlib import Path
from tqdm import tqdm

from ....curation.stac import STACDataFrame


class IngestSTAC:
    def __init__(self, repo, ingest_file, allowed_extensions, logger):
        self.repo = repo
        self.ingest_file = ingest_file
        self.allowed_extensions = allowed_extensions
        self.logger = logger if logger else print

    class Inputs(BaseModel):
        stac_catalog: Path
        user: dict

    class Outputs(BaseModel):
        dataset: dict

    def __call__(self, inputs: Inputs) -> Outputs:
        # retrieve the user's geodb credentials
        # creds, error = self.repo.retrieve_credentials(inputs.user["id_token"])
        # self.validate_credentials(creds)
        # load the STAC catalog as a STACsetFrame
        self.logger("Loading STAC catalog...")
        df = STACDataFrame.from_stac_file(inputs.stac_catalog)
        catalog = df[df["type"] == "Catalog"]
        assert len(catalog) == 1, "STAC catalog must have exactly one root catalog"
        dataset_name = catalog.id.iloc[0]
        # create dataset
        data, error = self.repo.create_stac_dataset(
            dataset_name, inputs.user["id_token"]
        )
        if error:
            data, error2 = self.repo.retrieve_dataset(dataset_name)
            if error2:
                raise Exception(error)
            if data["uid"] != inputs.user["sub"]:
                raise Exception("Dataset already exists.")
            dataset_id = data["id"]
            # TODO: put size to 0 or else will add up
        else:
            dataset_id = data["dataset_id"]
        # TODO: check that we can ingest in geodb
        # upload all assets to EOTDL
        self.logger("Uploading assets...")
        df2 = df.dropna(subset=["assets"])
        for row in tqdm(df2.iterrows(), total=len(df2)):
            # for asset in df.assets.dropna().values[:10]:
            try:
                for k, v in row[1]["assets"].items():
                    data = self.ingest_file(
                        v["href"],
                        dataset_id,
                        self.logger,
                        self.allowed_extensions + [".tif", ".tiff", ".jpg"],
                        verbose=False,
                        root=inputs.stac_catalog.parent.parent,  # esto será siempre así en STAC?
                    )
                    file_url = f"{self.repo.url}datasets/{data['dataset_id']}/download/{data['file_name']}"
                    df.loc[row[0], "assets"][k]["href"] = file_url
            except Exception as e:
                self.logger(f"Error uploading asset {row[0]}: {e}")
                break
        # ingest the STAC catalog into geodb
        self.logger("Ingesting STAC catalog...")
        data, error = self.repo.ingest_stac(
            json.loads(df.to_json()), dataset_id, inputs.user["id_token"]
        )
        if error:
            # TODO: delete all assets that were uploaded
            raise Exception(error)
        self.logger("Done")
        return self.Outputs(dataset=data)



"""
Data access module for eotdl package.
"""

from .download import download_sentinel_imagery, search_and_download_sentinel_imagery
from .search import search_sentinel_imagery
from .sentinelhub.parameters import SUPPORTED_SENSORS



"""
Module for STAC extent
"""

from datetime import datetime
from typing import List

import pystac

import rasterio

from ...tools import get_item_metadata


def get_dem_temporal_interval() -> pystac.TemporalExtent:
    """
    Get a temporal interval for DEM data
    """
    min_date = datetime.strptime("2011-01-01", "%Y-%m-%d")
    max_date = datetime.strptime("2015-01-07", "%Y-%m-%d")

    return pystac.TemporalExtent([(min_date, max_date)])


def get_unknow_temporal_interval() -> pystac.TemporalExtent:
    """
    Get an unknown temporal interval
    """
    min_date = datetime.strptime("2000-01-01", "%Y-%m-%d")
    max_date = datetime.strptime("2023-12-31", "%Y-%m-%d")

    return pystac.TemporalExtent([(min_date, max_date)])


def get_unknow_extent() -> pystac.Extent:
    """
    Get an unknown extent
    """
    return pystac.Extent(
        spatial=pystac.SpatialExtent([[0, 0, 0, 0]]),
        temporal=pystac.TemporalExtent(
            [
                (
                    datetime.strptime("2000-01-01", "%Y-%m-%d"),
                    datetime.strptime("2023-12-31", "%Y-%m-%d"),
                )
            ]
        ),
    )


def get_collection_extent(rasters: List[str]) -> pystac.Extent:
    """
    Get the extent of a collection

    :param rasters: list of rasters
    """
    # Get the spatial extent of the collection
    spatial_extent = get_collection_spatial_extent(rasters)
    # Get the temporal interval of the collection
    temporal_interval = get_collection_temporal_interval(rasters)
    # Create the Extent object
    extent = pystac.Extent(spatial=spatial_extent, temporal=temporal_interval)

    return extent


def get_collection_spatial_extent(rasters: List[str]) -> pystac.SpatialExtent:
    """
    Get the spatial extent of a collection

    :param path: path to the directory
    """
    # Get the bounding boxes of all the given rasters
    bboxes = []
    for raster in rasters:
        with rasterio.open(raster) as ds:
            bounds = ds.bounds
            dst_crs = "EPSG:4326"
            try:
                left, bottom, right, top = rasterio.warp.transform_bounds(
                    ds.crs, dst_crs, *bounds
                )
                bbox = [left, bottom, right, top]
            except rasterio.errors.CRSError:
                spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
                return spatial_extent
            bboxes.append(bbox)
    # Get the minimum and maximum values of the bounding boxes
    try:
        left = min(bbox[0] for bbox in bboxes)
        bottom = min(bbox[1] for bbox in bboxes)
        right = max(bbox[2] for bbox in bboxes)
        top = max(bbox[3] for bbox in bboxes)
        spatial_extent = pystac.SpatialExtent([[left, bottom, right, top]])
    except ValueError:
        spatial_extent = pystac.SpatialExtent([[0, 0, 0, 0]])
    finally:
        return spatial_extent


def get_collection_temporal_interval(rasters: List[str]) -> pystac.TemporalExtent:
    """
    Get the temporal interval of a collection

    :param path: path to the directory
    """
    # Get all the metadata.json files in the directory of all the given rasters
    metadata_jsons = []
    for raster in rasters:
        metadata_json = get_item_metadata(raster)
        if metadata_json:
            metadata_jsons.append(metadata_json)

    if not metadata_jsons:
        return (
            get_unknow_temporal_interval()
        )  # If there is no metadata, set a generic temporal interval

    # Get the temporal interval of every metadata.json file and the type of the data
    data_types = []
    temporal_intervals = []
    for metadata in metadata_jsons:
        # Append the temporal interval to the list as a datetime object
        temporal_intervals.append(metadata["acquisition-date"]) if metadata[
            "acquisition-date"
        ] else None
        # Append the data type to the list
        data_types.append(metadata["type"]) if metadata["type"] else None

    if temporal_intervals:
        min_date, max_date = None, None
        try:
            # Get the minimum and maximum values of the temporal intervals
            min_date = min(
                datetime.strptime(interval, "%Y-%m-%d")
                for interval in temporal_intervals
            )
            max_date = max(
                datetime.strptime(interval, "%Y-%m-%d")
                for interval in temporal_intervals
            )
        except ValueError:
            min_date = datetime.strptime("2000-01-01", "%Y-%m-%d")
            max_date = datetime.strptime("2023-12-31", "%Y-%m-%d")
        finally:
            # Create the temporal interval
            return pystac.TemporalExtent([(min_date, max_date)])
    else:
        # Check if the collection is composed by DEM data. If not, set a generic temporal interval
        if (
            set(data_types) == {"dem"}
            or set(data_types) == {"DEM"}
            or set(data_types) == {"dem", "DEM"}
        ):
            return get_dem_temporal_interval()
        else:
            return get_unknow_temporal_interval()



"""
Evalscripts for Sentinel Hub requests
"""


class EvalScripts:
    """
    Class that defines the needed Sentinel Hub evalscripts
    """

    SENTINEL_1 = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["VH", "VV"]
                        }],
                        output: {
                            id: "default",
                            bands: 2
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.VH, sample.VV];
                }
                """

    SENTINEL_2_L1C = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["B01",
                                    "B02",
                                    "B03",
                                    "B04",
                                    "B05",
                                    "B06",
                                    "B07",
                                    "B08",
                                    "B8A",
                                    "B09",
                                    "B10",
                                    "B11",
                                    "B12"],
                            units: "DN"
                        }],
                        output: {
                            bands: 13,
                            sampleType: "INT16"
                        }
                    };
                }

                function updateOutputMetadata(scenes, inputMetadata, outputMetadata) {
                    outputMetadata.userData = {"scenes" : scenes};
                }

                function evaluatePixel(sample) {
                    return [sample.B01,
                            sample.B02,
                            sample.B03,
                            sample.B04,
                            sample.B05,
                            sample.B06,
                            sample.B07,
                            sample.B08,
                            sample.B8A,
                            sample.B09,
                            sample.B10,
                            sample.B11,
                            sample.B12];
                }
                """

    SENTINEL_2_L2A = """
                //VERSION=3
                function setup() {
                    return {
                        input: [{
                            bands: ["B01",
                                    "B02",
                                    "B03",
                                    "B04",
                                    "B05",
                                    "B06",
                                    "B07",
                                    "B08",
                                    "B8A",
                                    "B09",
                                    "B11",
                                    "B12"],
                            units: "DN"
                        }],
                        output: {
                            bands: 12,
                            sampleType: "INT16"
                        }
                    };
                }

                function evaluatePixel(sample) {
                    return [sample.B01,
                            sample.B02,
                            sample.B03,
                            sample.B04,
                            sample.B05,
                            sample.B06,
                            sample.B07,
                            sample.B08,
                            sample.B8A,
                            sample.B09,
                            sample.B11,
                            sample.B12];
                }
                """

    DEM = """
        //VERSION=3

        function setup() {
            return {
                input: ["DEM"],
                output: { id: "default",
                        bands: 1,
                        sampleType: SampleType.FLOAT32
                },
            }
        }

        function evaluatePixel(sample) {
            return [sample.DEM]
        }
        """



"""
Module for the STAC label extension base object
"""

from typing import List, Union

import pystac
from pystac.extensions.label import (
    LabelClasses,
    LabelExtension,
    SummariesLabelExtension,
)

from ..base import STACExtensionObject


class LabelExtensionObject(STACExtensionObject):
    """
    STAC Label extension base object in EOTDL
    """
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def generate_stac_labels(cls) -> None:
        """
        Generate a labels collection from a STAC dataframe.
        """
        return

    def add_extension_to_item(
        self,
        obj: pystac.Item,
        label_description: str,
        label_type: str,
        label_names: List[str],
        label_classes: List[str],
        **kwargs
    ) -> Union[pystac.Item, pystac.Asset]:
        """
        Add the extension to the given object

        :param obj: object to add the extension
        :param label_description: label description
        :param label_type: label type
        :param label_names: list of label names
        :param label_classes: list of label classes of the item
        :param kwargs: optional arguments
            :param kwargs.label_properties: list of label properties
            :param kwargs.label_methods: list of label methods
            :param kwargs.label_tasks: list of label tasks

        :return: the item with the label extension
        """
        label_item = pystac.Item(
            id=obj.id,
            geometry=obj.geometry,
            bbox=obj.bbox,
            properties={},
            datetime=obj.datetime,
        )

        # Add the label extension to the item
        LabelExtension.add_to(label_item)

        # Access the label extension
        label_ext = LabelExtension.ext(label_item)

        # Add the label classes
        for name, classes in zip(label_names, label_classes):
            label_classes = LabelClasses.create(
                name=name,
                classes=classes,
            )
            label_ext.label_classes = [label_classes]

        # Add the label description
        label_ext.label_description = label_description
        # Add the label type
        label_ext.label_type = label_type
        # Add the label properties, if any
        label_ext.label_properties = (
            kwargs.get("label_properties")
            if kwargs.get("label_properties", None)
            else label_names
        )
        # Add the label methods, if any
        label_ext.label_methods = (
            kwargs.get("label_methods") if kwargs.get("label_methods", None) else None
        )
        # Add the label tasks, if any
        label_ext.label_tasks = (
            kwargs.get("label_tasks") if kwargs.get("label_tasks", None) else None
        )
        # Add the source
        label_ext.add_source(obj)

        return label_item

    def add_extension_to_collection(
        self,
        obj: pystac.Collection,
        label_names: List[str],
        label_classes: List[Union[list, tuple]],
        label_type: str,
    ) -> None:
        """
        Add the label extension to the given collection

        :param obj: object to add the extension
        :param label_names: list of label names
        :param label_classes: list of label classes
        :param label_type: label type
        """
        LabelExtension.add_to(obj)

        # Add the label extension to the collection
        label_ext = SummariesLabelExtension(obj)

        # Add the label classes
        for name, classes in zip(label_names, label_classes):
            label_classes = LabelClasses.create(
                name=name,
                classes=classes,
            )
            label_ext.label_classes = [label_classes]

        # Add the label type
        label_ext.label_type = label_type

    def add_geojson_to_items(self) -> None:
        """
        Add a GeoJSON FeatureCollection to every label item, as recommended by the spec
        https://github.com/stac-extensions/label#assets
        """
        return



from pathlib import Path
import yaml
from tqdm import tqdm
import json

from ..auth import with_auth
from .metadata import Metadata
from ..repos import DatasetsAPIRepo, FilesAPIRepo
from ..files import ingest_files, create_new_version
from ..curation.stac import STACDataFrame
from ..shared import calculate_checksum


def ingest_dataset(path, verbose=False, logger=print):
    path = Path(path)
    if not path.is_dir():
        raise Exception("Path must be a folder")
    if "catalog.json" in [f.name for f in path.iterdir()]:
        return ingest_stac(path / "catalog.json", logger)
    return ingest_folder(path, verbose, logger)


def retrieve_dataset(metadata, user):
    repo = DatasetsAPIRepo()
    data, error = repo.retrieve_dataset(metadata.name)
    # print(data, error)
    if data and data["uid"] != user["sub"]:
        raise Exception("Dataset already exists.")
    if error and error == "Dataset doesn't exist":
        # create dataset
        data, error = repo.create_dataset(metadata.dict(), user["id_token"])
        # print(data, error)
        if error:
            raise Exception(error)
        data["id"] = data["dataset_id"]
    return data["id"]


@with_auth
def ingest_folder(folder, verbose=False, logger=print, user=None):
    repo = DatasetsAPIRepo()
    # load metadata
    metadata = yaml.safe_load(open(folder.joinpath("metadata.yml"), "r").read()) or {}
    metadata = Metadata(**metadata)
    # retrieve dataset (create if doesn't exist)
    dataset_id = retrieve_dataset(metadata, user)
    # ingest files
    return ingest_files(
        repo, dataset_id, folder, verbose, logger, user, endpoint="datasets"
    )


def retrieve_stac_dataset(dataset_name, user):
    repo = DatasetsAPIRepo()
    data, error = repo.retrieve_dataset(dataset_name)
    # print(data, error)
    if data and data["uid"] != user["sub"]:
        raise Exception("Dataset already exists.")
    if error and error == "Dataset doesn't exist":
        # create dataset
        data, error = repo.create_stac_dataset(dataset_name, user["id_token"])
        # print(data, error)
        if error:
            raise Exception(error)
        data["id"] = data["dataset_id"]
    return data["id"]


@with_auth
def ingest_stac(stac_catalog, logger=None, user=None):
    repo, files_repo = DatasetsAPIRepo(), FilesAPIRepo()
    # load catalog
    logger("Loading STAC catalog...")
    df = STACDataFrame.from_stac_file(stac_catalog)
    catalog = df[df["type"] == "Catalog"]
    assert len(catalog) == 1, "STAC catalog must have exactly one root catalog"
    dataset_name = catalog.id.iloc[0]
    # retrieve dataset (create if doesn't exist)
    dataset_id = retrieve_stac_dataset(dataset_name, user)
    # create new version
    version = create_new_version(repo, dataset_id, user)
    logger("New version created, version: " + str(version))
    df2 = df.dropna(subset=["assets"])
    for row in tqdm(df2.iterrows(), total=len(df2)):
        # for asset in df.assets.dropna().values[:10]:
        try:
            for k, v in row[1]["assets"].items():
                data, error = files_repo.ingest_file(
                    v["href"],
                    dataset_id,
                    user["id_token"],
                    calculate_checksum(v["href"]),  # is always absolute?
                    "datasets",
                    version,
                )
                if error:
                    raise Exception(error)
                file_url = f"{repo.url}datasets/{data['dataset_id']}/download/{data['filename']}"
                df.loc[row[0], "assets"][k]["href"] = file_url
        except Exception as e:
            logger(f"Error uploading asset {row[0]}: {e}")
            break
    # ingest the STAC catalog into geodb
    logger("Ingesting STAC catalog...")
    data, error = repo.ingest_stac(
        json.loads(df.to_json()), dataset_id, user["id_token"]
    )
    if error:
        # TODO: delete all assets that were uploaded
        raise Exception(error)
    logger("Done")
    return



import requests
import geopandas as gpd

from ..repos import APIRepo


class DatasetsAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def retrieve_datasets(self, name, limit):
        url = self.url + "datasets"
        if name is not None:
            url += "?match=" + name
        if limit is not None:
            if name is None:
                url += "?limit=" + str(limit)
            else:
                url += "&limit=" + str(limit)
        response = requests.get(url)
        return self.format_response(response)

    def create_dataset(self, metadata, id_token):
        response = requests.post(
            self.url + "datasets",
            json=metadata,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)

    def retrieve_dataset(self, name):
        response = requests.get(self.url + "datasets?name=" + name)
        return self.format_response(response)

    def create_version(self, dataset_id, id_token):
        response = requests.post(
            self.url + "datasets/version/" + dataset_id,
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)

    def create_stac_dataset(self, name, id_token):
        response = requests.post(
            self.url + "datasets/stac",
            json={"name": name},
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)

    def ingest_stac(self, stac_json, dataset_id, id_token):
        response = requests.put(
            self.url + f"datasets/stac/{dataset_id}",
            json={"stac": stac_json},
            headers={"Authorization": "Bearer " + id_token},
        )
        return self.format_response(response)

    def download_stac(self, dataset_id, id_token):
        url = self.url + "datasets/" + dataset_id + "/download"
        headers = {"Authorization": "Bearer " + id_token}
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            return None, response.json()["detail"]
        return gpd.GeoDataFrame.from_features(response.json()["features"]), None

    # def update_dataset(self, name, path, id_token, checksum):
    #     # check that dataset exists
    #     data, error = self.retrieve_dataset(name)
    #     if error:
    #         return None, error
    #     # first call to get upload id
    #     dataset_id = data["id"]
    #     url = self.url + f"datasets/chunk/{dataset_id}?checksum={checksum}"
    #     response = requests.get(url, headers={"Authorization": "Bearer " + id_token})
    #     if response.status_code != 200:
    #         return None, response.json()["detail"]
    #     data = response.json()
    #     _, upload_id, parts = data["dataset_id"], data["upload_id"], data["parts"]
    #     # assert dataset_id is None
    #     content_path = os.path.abspath(path)
    #     content_size = os.stat(content_path).st_size
    #     url = self.url + "datasets/chunk"
    #     chunk_size = 1024 * 1024 * 100  # 100 MiB
    #     total_chunks = content_size // chunk_size
    #     headers = {
    #         "Authorization": "Bearer " + id_token,
    #         "Upload-Id": upload_id,
    #         "Dataset-Id": dataset_id,
    #     }
    #     # upload chunks sequentially
    #     pbar = tqdm(
    #         self.read_in_chunks(open(content_path, "rb"), chunk_size),
    #         total=total_chunks,
    #     )
    #     index = 0
    #     for chunk in pbar:
    #         offset = index + len(chunk)
    #         part = index // chunk_size + 1
    #         index = offset
    #         if part not in parts:
    #             headers["Part-Number"] = str(part)
    #             file = {"file": chunk}
    #             r = requests.post(url, files=file, headers=headers)
    #             if r.status_code != 200:
    #                 return None, r.json()["detail"]
    #         pbar.set_description(
    #             "{:.2f}/{:.2f} MB".format(
    #                 offset / 1024 / 1024, content_size / 1024 / 1024
    #             )
    #         )
    #     pbar.close()
    #     # complete upload
    #     url = self.url + "datasets/complete"
    #     r = requests.post(
    #         url,
    #         json={"checksum": checksum},
    #         headers={
    #             "Authorization": "Bearer " + id_token,
    #             "Upload-Id": upload_id,
    #             "Dataset-Id": dataset_id,
    #         },
    #     )
    #     if r.status_code != 200:
    #         return None, r.json()["detail"]
    #     return r.json(), None
